// The program used in the paper of RC2024.

// Lyngby, T., Nylandsted, R.R., Glueck, R., Yokoyama, T. (2024). Towards Clean Reversible Lossless Compression. In: Mogensen, T.AE., Mikulski, L. (eds) Reversible Computation. RC 2024. Lecture Notes in Computer Science, vol 14680, 94-102, Springer. https://doi.org/10.1007/978-3-031-62076-8_7

// BWT_O_n3_rc2024.j authored by Therese Lyngby and Rasmus Ross Nylandsted (June 2023)
//                   updated by Tetsuo Yokoyama (May 2024)
// Time O(n^3), Space O(n)
// It is tested on janus-playground-180405: https://tetsuo.jp/janus-playground-180405/#e5498475

// Procedures that manipulate ''strings'' ---------
// Strings are represented as zero ended integer arrays
// We allow the characters '0' -- '9', where 0 is a special end-of-string character
// The first index of the array contains the length of the string

// Takes an array and its size n.
// The array is filled with the numbers from 0 to n - 1
procedure fill_array_up_to_n(int array[], int n)
    if n >= 0 then
        iterate int i = 0 to n-1
            array[i] += i
        end
    fi n >= 0

// Checks whether offsets[matchindex] is lexicographically smaller than the index immediatly following
// Result is set to non-zero if yes
// Result should initially be 0
// If matchindex is the last index it is sorted by default
procedure is_lexicographically_sorted(int in[], int offsets[], int matchindex, int result)

    // If last element matrix[matchindex] is sorted by default
    if matchindex = size(in) - 1 || size(in) = 0 then
        result += 1
    else
        local int i = 0 // iterates through the two strings
        local int n = size(in)
        local int offset1 = (offsets[matchindex])
        local int offset2 = (offsets[matchindex + 1])

        // compare matchindex and matchindex + 1 character by character
        from i = 0 do
            i += 1
            if in[offset1 % n] < in[offset2 % n] || i = n then // one larger than word aka out of bounds
                result += 1
            fi result = 1
        loop
            offset1 += 1
            offset2 += 1
        until result = 1 || in[offset1 % n] > in[offset2 % n]

        // reset i
        from i = n || in[offset1 % n] != in[offset2 % n] do
            i -= 1
        loop
            offset1 -= 1
            offset2 -= 1
        until i = 0

        delocal int offset2 = (offsets[matchindex + 1])
        delocal int offset1 = (offsets[matchindex])
        delocal int n = size(in)
        delocal int i = 0
    fi matchindex = size(in) - 1 || size(in) = 0

// Searches through array to find term and updates matchindex to be
// the first index containing term. Array does NOT have string format
// matchindex should initially be zero-cleared, and is set to -1 if no match
procedure search(int term, int array[], int matchindex)
    from matchindex = 0 loop
        matchindex += 1
    until (matchindex = size(array)) || (array[matchindex] = term)

    // sets matchindex to -1 if no match is found
    if matchindex = size(array) then
        matchindex -= (size(array) + 1)
    fi matchindex < 0

// swaps the offsets such that they are sorted from smallest to largest
// Assumes the offsets are initially in lexicographical order
procedure sort(int in[], int array[])
    iterate int i = 0 to size(array) - 1
        local int matchindex = 0
        local int is_sorted = 0

        // Find where i is in array
        call search(i, array, matchindex)

        call is_lexicographically_sorted(in, array, matchindex, is_sorted)

        // move i from match index to its proper position
        from is_sorted = 1 do
            uncall is_lexicographically_sorted(in, array, matchindex, is_sorted)
        loop
            array[matchindex] <=> array[matchindex-1]
            matchindex -= 1
            call is_lexicographically_sorted(in, array, matchindex, is_sorted)
        until matchindex = i

        delocal int is_sorted = 0
        delocal int matchindex = i
    end

// Takes input, an offset array, where the offsets are sorted lexicographically
// and an index of the original word, and 'swaps' into an offset
// should zero-clear index upon execution
procedure unsort_lexicographically(int in[], int offsets[], int idx)
    // zero-clear idx
    iterate int i = 0 to size(offsets) - 1
        if offsets[i] = 0 then
            idx -= i
        fi offsets[i] = 0
    end

    call sort(in, offsets)

procedure calculate_c_and_p(int c[], int p[], int out[])
    // iterate through array p
    iterate int i = 0 to size(out) - 1
        p[i] += c[out[i]]
        c[out[i]] += 1
    end

    local int sum = 0

    // iterate through alphabet
    iterate int ch = 0 to 9
        sum += c[ch]

        // We want to calculate c[ch] = sum - c[ch]
        local int tmp = 0
        c[ch] <=> tmp
        c[ch] += sum - tmp
        delocal int tmp = sum - c[ch]
    end
    delocal int sum = size(out)

procedure calculate_l(int c[], int p[], int out[], int idx, int l[])
    local int n = size(l)
    l[n - 1] += idx
    iterate int j = n - 2 by -1 to 0
        l[j]  += p[l[j+1]] + c[out[l[j+1]]]
    end

    delocal int n = size(l)

procedure gen_in(int out[], int idx, int in[])
        local int c[10], int p[size(in)], int l[size(in)]

        call calculate_c_and_p(c, p, out)
        call calculate_l(c, p, out, idx, l)

        iterate int j = size(in)-1 by -1 to 0
            in[j] += out[l[j]]
        end

        uncall calculate_l(c, p, out, idx, l)
        uncall calculate_c_and_p(c, p, out)

        delocal int c[10], int p[size(in)], int l[size(in)]

procedure bwt'(int in[], int out[], int idx, int offsets[])
    // fills an array with possible offsets into in
    call fill_array_up_to_n(offsets, size(in))

    // should set idx to the location of in place after sorting
    uncall unsort_lexicographically(in, offsets, idx)

procedure build_matrix(int out[], int idx, int offsets[], int in[])
    call gen_in(out, idx, in)
    local int idx' = 0, int n = size(in)
    call bwt'(in, out, idx', offsets)
    iterate int i = 0 to n - 1
        out[i] -= in[(offsets[i] + n - 1) % n]
    end
    delocal int idx' = idx, int n = size(in)

// takes an input string, an empty string to contain the output and an idx to
// contain the original input of the matrix, should initially be 0
procedure rev_bwt(int in[], int out[], int idx)
    local int offsets[size(in)]
    call bwt'(in, out, idx, offsets)
    uncall build_matrix(out, idx, offsets, in)
    delocal int offsets[size(in)]

/// test of 'rev_bwt' (general case)
procedure main()
    int in[] = {2, 1, 3, 1, 3, 1}
    int out[size(in)]
    int idx = 0

    call rev_bwt(in, out, idx)
    //uncall rev_bwt(in, out, idx)
/// Expected result: in = {0, 0, 0, 0, 0, 0}
///                  out = {3, 3, 2, 1, 1, 1}
///                  idx = 3
