// The program used in the paper of RC2024.

// Lyngby, T., Nylandsted, R.R., Glueck, R., Yokoyama, T. (2024). Towards Clean Reversible Lossless Compression. In: Mogensen, T.AE., Mikulski, L. (eds) Reversible Computation. RC 2024. Lecture Notes in Computer Science, vol 14680, 94-102, Springer. https://doi.org/10.1007/978-3-031-62076-8_7

// LZW_rc2024.j authored by Therese Lyngby and Rasmus Ross Nylandsted (June 2023)
//              updated by Therese Lyngby (May 2024)
// Time O(n), Space O(n)
// It is tested on janus-playground-180405: https://tetsuo.jp/janus-playground-180405/#bd58d880

// Procedures that manipulate ''strings'' ---------
// Strings are represented as zero ended integer arrays
// We allow the characters 0 -- 9, where 0 is a special end-of-string character
// The first index of the array contains the length of the string

// copies the content of str1 to str2
procedure copy_str(int str1[], int str2[])
    iterate int i = 0 to str1[0]
        str2[i] += str1[i]
    end

// --- Helper functions for Reversible LZW ---

// Initialize dict with the characters 0 -- 9
// d is the dictionary
// n is set to the number of entries in d
procedure init_dict(int d[][])
    iterate int i = 0 to 9
        d[i][0] += 1 // sets the size of the entry
        d[i][1] += i // sets the char
        // next character is already 0, so end-of-string is already set
    end


// compares str1 and str2, i is used to iterate through the strings
// Stops if the length of the strings doesn't match or if the strings differ
procedure compare(int str1[], int str2[], int i)
    // compares the size of the two strings
    if str1[0] = str2[0] then
        local int str1Len = str1[0]
        from i = 1 loop
            i += 1
        until (str1[i] != str2[i]) || (i = str1Len)
        delocal int str1Len = str1[0]
    fi str1[0] = str2[0]


// Finds the code in d, which correspond to newStr
// If no such entry exists matchIdx is set to -1
// matchIdx should also initially be -1
// n is the number of entries in d
procedure match(int newStr[], int matchIdx, int d[][], int n)
    local int i = 1
    local int curDictStr[size(newStr)]

    matchIdx += 1 // set matchIdx to zero

    from matchIdx = 0 do
        call copy_str(d[matchIdx], curDictStr)
        call compare(curDictStr, newStr, i)
    loop
        uncall compare(curDictStr, newStr, i)
        uncall copy_str(d[matchIdx], curDictStr)
        matchIdx += 1
    until ((i = newStr[0]) && (curDictStr[i] = newStr[i])) || (matchIdx = size(d) - 1)

    // if we stopped bc we got to the end of the dictionary
    if (curDictStr[i] != newStr[i]) || (i != newStr[0]) then
        uncall compare(curDictStr, newStr, i)
        uncall copy_str(d[matchIdx], curDictStr)
        matchIdx -= size(d)
    // stopped bc we found a match
    else
        // cannot be done before if bc if needs i, which is reset
        uncall compare(curDictStr, newStr, i)
        uncall copy_str(d[matchIdx], curDictStr)
    fi matchIdx = -1

    delocal int curDictStr[size(newStr)]
    delocal int i = 1


// combines the entry in d at fstCode with the first character of the entry in
// d at sndCode, and stores the result in newStr.
// n is the number of entries in d
// we assume newStr is big enough to contain the combined strings,
// and that newStr is initially empty
// fstCode is zero-cleared during execution
procedure combine(int fstCode, int sndCode, int newStr[], int d[][], int n)
    local int len = 0, int tmp[size(newStr)]

    // copies over the entry in d at fstCode
    call copy_str(d[fstCode], newStr)
    len += newStr[0]

    // zero-clear fstCode
    uncall match(newStr, fstCode, d, n) // sets fstCode to -1
    fstCode += 1 // now we have zero-cleared fstIdx

    // Gets the first character of the second dictionary entry
    call copy_str(d[sndCode], tmp)
    // adds the first character of the dictionary entry at sndCode
    newStr[len + 1] += tmp[1]
    len -= newStr[0] // zero-clear length
    newStr[0] += 1 // update length field of newStr
    uncall copy_str(d[sndCode], tmp) // resets tmp

    delocal int len = 0, int tmp[size(newStr)]

// Adds str to the d. n is the number of entries in d
// str is NOT zero-cleared
procedure add_to_dict(int str[], int d[][], int n)
    iterate int i = 0 to str[0]
        d[n][i] += str[i]
    end
    n += 1

// Finds the next subStr in in starting at index i, which is not already in d
// subStr is added to d, i is moved to the last character of subStr
// subStr of in is zero-cleared, curCode is the entry of the longest substr
// already in d
procedure find_str(int in[], int i, int d[][], int n, int curCode)
    local int matchIdx = -1
    local int curChar = 0
    local int newStr[size(in)]

    curCode <=> in[i]

    // 10 is the language size, has been hard coded for now
    from curCode < 10 do
        i += 1
        curChar += in[i]
        call combine(curCode, curChar, newStr, d, n)
        call match(newStr, matchIdx, d, n)

    loop
        in[i] -= curChar // zero-clear input

        // zero-clear curChar (identical to last string char)
        local int len = newStr[0]
        curChar -= newStr[len]
        delocal int len = newStr[0]

        curCode <=> matchIdx
        matchIdx -= 1   // match is set back to -1
        uncall copy_str(d[curCode], newStr) // resets newStr

    // will always stop if no more input, <something>0 is not in d
    until matchIdx = -1

    // only add new entry if dictionary is not full
    if n < size(d) then
        call add_to_dict(newStr, d, n) // Does not zero-clear newStr
    else
        n += 1 // we still count up number of entries
    fi n < size(d) + 1
    uncall combine(curCode, curChar, newStr, d, n)

    delocal int newStr[size(in)]
    delocal int curChar = in[i]
    delocal int matchIdx = -1


// Assumes d is big enough to contain another entry
// Adds str to d. str is zero-cleared
// n is the number of entries in d
procedure push_to_dict(int str[], int d[][], int n)
    local int strLen = str[0]
    iterate int i = 0 to strLen
        d[n][i] <=> str[i]
    end
    n += 1

    delocal int strLen = d[n - 1][0]


// Generates d from out, and calculates the len of the original in.
// n is the number of entries in d.
// inLen stores the size of the original in
// d inLen must initiallly be zero-cleared
procedure build_dict(int out[], int inLen, int d[][])
    call init_dict(d)
    local int j = 0, int n = 10, int curCode = out[j], int prevCode = 0, int newStr[size(out)+1]

    // 10 is the language size, has been hard coded for now
    from n = 10 loop
        j += 1
        inLen += d[curCode][0]
        prevCode <=> curCode
        curCode += out[j]

        // Handles KwKwK case
        if curCode = n then
            local int tmp = prevCode // local copy of prevCode
            call combine(prevCode, tmp, newStr, d, n)
            delocal int tmp = out[j - 1]

        // Normal case
        else
            call combine(prevCode, curCode, newStr, d, n)
        fi curCode = n

        //Check if room for more entries  in d
        if n < size(d) then
            call push_to_dict(newStr, d, n) // zero-clears newStr
        else
            uncall combine(prevCode, curCode, newStr, d, n)
	        n += 1
            prevCode -= out[j - 1] // zero-clear prevCode
        fi n < size(d) + 1
    until out[j] = 0

    delocal int j = size(out)-1, int n = 10+j, int curCode = 0, int prevCode = 0, int newStr[size(out)+1]

//Performs the main lzw compression, generating the auxillary data structure d
// in is the input. i an index into the input, should initially zero.
// out is the output, should initially be zero-cleared
// d is the dictionary, should initiallly be zero-cleared
// n the number of entries in d, should also initiallly be zero
procedure lzw' (int in[], int i, int out[], int d[][])
    call init_dict(d)

    local int j = 0, int n = 10, int curCode = 0
    from i = 0 loop
        call find_str(in, i, d, n, curCode)
        out[j] <=> curCode
        j += 1
    until in[i] = 0
    delocal int j = size(out)-1, int n = 10+j, int curCode = 0

// Implementation of reversible_lzw,
// in is the input and out stores the output
// out should initially be zero-cleared
procedure rev_lzw (int in[], int out[])
    local int i = 0, int d[24][17] //dict of fixed size
    call lzw'(in, i, out, d)
    uncall build_dict(out, i, d)
    delocal int i = 0, int d[24][17]

/// Test of 'rev_lzw' (Word "TANBANANAS", kwkwk case) T
procedure main()
    // Code: {T, A, N, B, A, N, A, N, A, S}
    int in[] = {1, 2, 3, 4, 2, 3, 2, 3, 2, 5, 0}
    int out[8]

    call rev_lzw(in, out)
    //uncall rev_lzw(in, out)
/// Expected Result: in[11] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
///                  out[8] = {1, 2, 3, 4, 11, 14, 5, 0}
